/*
Erik Herkenrath
begonnen: 24.Nov2010
beendet:  22.Jan2011

Funktion:
steuert bis zu 6 Lfter seperat ber PWM-Ausgnge an
jedem Lfter ist einer von 4 NTC-Temperatursensoren zuweisbar
zu jedem Lfter lsst sich die Anlauftemperatur T(min), Temp. fr max Drehzahl T(max) und maximal zu erreichender PWM-Wert PWM(max) einstellen
es sind 2 Profile gespeichert, welche mit dem Schalter "game-Mode" gewhlt werden

Daten:
C: ATmega168
F_CPU = 8 MHz (interner RC-Oszillator)
PWM-Frequenz: 31.25 kHz

Display			Pin
D0				B4
D1				B5
D2				B6
D3				B7
RS				D0
RW				D1
Enable			D2

Tasten			Pin
[1] hoch		D4
[4] links '-'	C5
[2] rechts '+'	D7
[3] runter		C4
game-Mode		B0	(LOW = silent, HIGH = game)
reset			C6

NTC/Temp.Sens	Pin
T1				C0
T2				C1
T3				C2
T4				C3

Lfter			Pin
Fan1 (OC0A)		D6
Fan2 (OC0B)		D5
Fan3 (OC1A)		B1
Fan4 (OC1B)		B2
Fan5 (OC2A)		B3
Fan6 (OC2B)		D3
*/

//////////////////
// Definitionen //
//////////////////

#define F_CPU 8000000UL		// 8 MHz CPU-Takt, max 32.7 ms delay!
#define PWMmin 140			// minimal mglicher PWM-Wert


////////////////////
// Header-Dateien //
////////////////////

#include <avr/io.h>			// Portpin & ATmega168 Definitionen
#include <util/delay.h>		// Verzgerungen
#include <avr/eeprom.h>		// Variablen im EEPROM speichern

#include "LCD/lcd.h"		// Ausgabe ber LCD-Display


////////////////////////////////////
// globale Variablendeklarationen //
////////////////////////////////////

uint8_t i = 0;						// Zhlvariable
uint8_t line = 0;					// aktuelle Displayzeile zum bearbeiten
uint8_t fan_show = 1;				// aktuell angezeigter Lfter
uint8_t fan_read = 1;				// aktuell ausgelesener Lfter
uint8_t sens_read = 1;				// aktuell ausgelesener Sensor
uint8_t mode = 0;					// silent-Modus: 0, game-Modus: 1
uint8_t button = 0;					// zuletzt gedrckte Taste (auer: mode)
uint8_t button_counter = 0;			// lnge des Tastendruckes (* 20 ms)
uint16_t adc_value = 0;				// Ergebnis des ADC
uint16_t Tmin = 0;					// zur PWM-Berechnung: T(min) als LSB-Wert
uint16_t Tmax = 0;					// zur PWM-Berechnung: T(max) als LSB-Wert
char buffer[4] = {0, 0, 0, 0};		// String zur Ausgabe von Zahlenwerten ber das Display
uint8_t temp_sens[5];				// zuletzt vom jeweiligen Sensor ausgelesene Temperatur (lediglich zur Displayausgabe)
uint8_t settings[2][7][4];			// 3-dimensionales Array: [mode][Lfter][Zeile]
										// [Zeile]:  0:Sensor, 1:t_min, 2:t_max, 3:pwm_max

// EEPROM-Variablen (werden zu Beginn in RAM eingelesen)
uint8_t ee_settings[2][7][4] EEMEM =	// 3-dimensionales Array: [mode][Lfter][Zeile]
{										// Zeile:  0:Sensor, 1:t_min, 2:t_max, 3:pwm_max
 {							// silent-mode
  { 0, 0, 0, 0},			// leer (kein Lfter0)
  { 1, 60, 70, 200 },		// Lfter1
  { 1, 60, 70, 200 },		// Lfter2
  { 2, 60, 70, 200 },		// Lfter3
  { 2, 60, 70, 200 },		// Lfter4
  { 3, 60, 70, 200 },		// Lfter5
  { 4, 60, 70, 200 }		// Lfter6
 },
 {							// game-mode
  { 0, 0, 0, 0},			// leer (kein Lfter0)
  { 1, 60, 70, 200 },		// Lfter1
  { 1, 60, 70, 200 },		// Lfter2
  { 2, 60, 70, 200 },		// Lfter3
  { 2, 60, 70, 200 },		// Lfter4
  { 3, 60, 70, 200 },		// Lfter5
  { 4, 60, 70, 200 }		// Lfter6
 }
};

volatile uint8_t * const PWM_register[7] =	// pointer auf die jeweiligen Register zur PWM-Steuerung (konstanter pointer)
{
  0,		// kein Fan0
  &OCR0A,	// Fan1
  &OCR0B,	// Fan2
  &OCR1AL,	// Fan3 (16-Bit-Register, daher nur L = unteres Register)
  &OCR1BL,	// Fan4
  &OCR2A,	// Fan5
  &OCR2B	// Fan6
};


////////////////
// Funktionen //
////////////////

void my_utoa3(const uint8_t v, char *s)	// wandelt uint8_t in 3-stellige Zahl mit anfhrenden Leerzeichen um
{
	uint8_t temp1 = v;			// darzustellende zahl abzglich bereits dargestellter Dezimalstellen
	uint8_t temp2 = 0;			// eine Dezimalstelle
	temp2 = temp1 / 100;
	if(temp2 > 0)
		s[0] = (temp2 + '0');	// hunderter
	else
		s[0] = ' ';				// wenn <100 dann Leerzeichen
	temp1 %= 100;				// Rest bei Division druch 100
	temp2 = temp1 / 10;
	if(temp2 > 0)
		s[1] = (temp2 + '0');	// zehner
	else if(s[0] == ' ')		// wenn <10 dann ' ', sonst '0'
		s[1] = ' ';
	else
		s[1] = '0';
	temp1 %= 10;				// Rest bei Division druch 10
	s[2] = (temp1 + '0');		// einer
}

void my_utoa3_T(const uint8_t v, char *s)	// wandelt temperaturwert in 3-stellige Zahl mit anfhrenden Leerzeichen um
{											// Temperatur als unsigned gespeichert: wert=0 entspricht -30C, wert=30 entspricht 0C	
	uint8_t temp1 = 0;				// darzustellende zahl abzglich bereits dargestellter Dezimalstellen
	uint8_t temp2 = 0;				// eine Dezimalstelle

	if(v == 0)						// kein Sensor angeschlossen
	{
		s[0] = ' ';
		s[1] = '-';
		s[2] = ' ';
	}
	else
	{
		if(v >29)					// Temperatur positiv
		{
			temp1 = v - 30;
			temp2 = temp1 / 100;
			if(temp2 > 0)
				s[0] = (temp2 + '0'); // hunderter
			else
				s[0] = ' ';			// wenn <100 dann Leerzeichen
			temp1 %= 100;			// Rest bei Division druch 10
		}
		else						// Temperatur negativ
		{
			temp1 = 30 - v;
			s[0] = '-';
		}	
		temp2 = temp1 / 10;
		if(temp2 > 0)
			s[1] = (temp2 + '0');	// zehner
		else if(s[0] == ' ' || s[0] == '-')	// wenn <10 dann ' ', sonst '0'
			s[1] = ' ';
		else
			s[1] = '0';
		temp1 %= 10;				// Rest bei Division druch 100
		s[2] = (temp1 + '0');		// einer
	}
}

void quick_refresh_screen()		// schreibt alle dynamischen Zahlenwerte neu (T & PWM)
{
	// Temperatur
	my_utoa3_T(temp_sens[(settings[mode][fan_show][0])], buffer);
	lcd_gotoxy(6, 0);
	lcd_puts(buffer);
	// PWM-Wert
	my_utoa3(*PWM_register[fan_show], buffer);	
	lcd_gotoxy(16, 3);
	lcd_puts(buffer);
}

void refresh_screen()			// schreibt alle statischen Zahlenwerte neu
{
	// Sensor
	lcd_gotoxy(3, 0);
	lcd_putc(settings[mode][fan_show][0] + 48);
	// Lfter
	lcd_gotoxy(19, 0);
	lcd_putc(fan_show + 48);
	// t_ein
	my_utoa3_T(settings[mode][fan_show][1], buffer);
	lcd_gotoxy(11,1);
	lcd_puts(buffer);
	// t_max
	my_utoa3_T(settings[mode][fan_show][2], buffer);
	lcd_gotoxy(11,2);
	lcd_puts(buffer);
	// pwm_max
	my_utoa3(settings[mode][fan_show][3], buffer);
	lcd_gotoxy(11,3);
	lcd_puts(buffer);
	// dynamische Werte
	quick_refresh_screen();
}

void print_screen()				// schreibt komplette Displaybeschriftung
{
	line = 0;
	lcd_clrscr();
	// 1. Zeile: "->T:- (  -C) Lfter-"
	lcd_puts("~T:- (  -C) Lfter-");
	// 2. Zeile: "T(ein) =  -C"
	lcd_gotoxy(2,1);
	lcd_puts("T(ein) =   -C");
	// 3. Zeile: "T(max) =  -C"
	lcd_gotoxy(2,2);
	lcd_puts("T(max) =   -C");
	// 4. Zeile: "PWM(max)=   - (  -)"
	lcd_gotoxy(1,3);
	lcd_puts("PWM(max)=   - (  -)");

	refresh_screen();
}


///////////////////
// main-Funktion //
///////////////////

int main()
{
	///////////////////////
	// Initialisierungen //
	///////////////////////

	// 6 PWM-Ausgnge: Pins als Ausgang definieren
	DDRD  = (1<<PIN3) | (1<<PIN5) | (1<<PIN6);
	DDRB  = (1<<PIN1) | (1<<PIN2) | (1<<PIN3);

	// Pullupwiderstnde fr Taster aktivieren
	PORTB = (1<<PIN0);
	PORTC = (1<<PIN4) | (1<<PIN5);
	PORTD = (1<<PIN4) | (1<<PIN7);

	// Register
	// Timer0
	TCCR0A  = (1<<COM0A1);		// non inverting PWM0A
	TCCR0A |= (1<<COM0B1);		// non inverting PWM0B
	TCCR0A |= (1<<WGM00) | (1<< WGM01);	// Fast PWM, TOP: 0xFF
	TCCR0B  = (1<<CS00);		// Prescaler = 1
	OCR0A = 0;					// alle Lfter auf 0/255 initialisieren
	OCR0B = 0;

	// Timer1
	TCCR1A  = (1<<COM1A1);		// non inverting PWM1A
	TCCR1A |= (1<<COM1B1);		// non inverting PWM1B
	TCCR1A |= (1<< WGM10);		// Fast PWM, TOP: 0x00FF
	TCCR1B  = (1<<WGM12);		// siehe: WGM10
	TCCR1B |= (1<<CS10);		// Prescaler = 1
	OCR1A = 0;
	OCR1B = 0;
	
	// Timer2
	TCCR2A  = (1<<COM2A1);		// non inverting PWM2A
	TCCR2A |= (1<<COM2B1);		// non inverting PWM2B
	TCCR2A |= (1<<WGM20) | (1<< WGM21);	// Fast PWM, TOP: 0xFF
	TCCR2B  = (1<<CS20);		// Prescaler = 1
	OCR2A = 0;
	OCR2B = 0;
	
	// ADC
	ADMUX = (1<<REFS0);			// Referenzspannung = AVCC (Kondensator an AREF!)
	DIDR0 = (1<<ADC0D) | (1<<ADC1D) | (1<<ADC2D) | (1<<ADC3D);	// Digitaleingnge der Pins 0-4 deaktivieren
	ADCSRA = (1<<ADEN);			// ADC einschalten
	ADCSRA |= (1<<ADPS1) | (1<<ADPS2);	// Prescaler 64 (bei 8 MHz = 125 kHz)

	// EEPROM
	eeprom_read_block(settings, ee_settings, 56);	// Einstellungen aus dem EEPROM laden

	// LCD-Display
	lcd_init(LCD_DISP_ON);
	// statische Zeichen
	print_screen();

	
	//////////////
	// Programm //
	//////////////

	// Endlosscheife
	while(1)
	{
		// Tastenentprellung
		if (!(PIND & (1<<PIN4)))			// Taste "hoch" gedrckt (Nr.1)
		{
			if(!(PINC & (1<<PIN4)))			// Taste "hoch" & "runter" gedrckt -> Einstellungen speichern! (Nr.5)
				button = 5;
			else
				button = 1;
			button_counter ++;
		}
		else if (!(PIND & (1<<PIN7)))		// Taste "rechts" gedrckt (Nr.2)
		{
			button = 2;
			button_counter ++;
		}
		else if (!(PINC & (1<<PIN4)))		// Taste "runter" gedrckt (Nr.3)
		{
			button = 3;
			button_counter ++;
		}
		else if (!(PINC & (1<<PIN5)))		// Taste "links" gedrckt (Nr.4)
		{
			button = 4;
			button_counter ++;
		}
		else if(button_counter != 0)		// keine Taste gedrckt: rcksetzen
			button_counter = 0;

		if(mode != (PINB & (1<<PIN0)) )		// Modus gendert?
		{
			mode = (PINB & (1<<PIN0));		// neuen Modus bernehmen
			refresh_screen();
		}


		// Tastenauswertung
		if(button_counter == 2)
		{
			switch(button)
			{
				case 1:							// hoch
					if(line == 0)					// oberste Zeile erreicht: Lfter wechseln
						if(fan_show > 1)
						{
							fan_show --;
							refresh_screen();
						}
						else
						{
							fan_show = 6;
							refresh_screen();
						}
					else							// sonst: eine Zeile hoch
					{
						lcd_gotoxy(0, line);
						lcd_putc(' ');
						line --;
						lcd_gotoxy(0, line);
						lcd_putc(126); // Pfeil
					}
					break;
				case 2:							// rechts (+)
					switch(line)
					{
						// Sensor
						case 0: if(settings[mode][fan_show][0] < 4)
									settings[mode][fan_show][0] ++; break;
						// T(ein)
						case 1: if(settings[mode][fan_show][1] < (settings[mode][fan_show][2] - 5))
									settings[mode][fan_show][1] += 5; break;
						// T(max)
						case 2: if(settings[mode][fan_show][2] < 100)
									settings[mode][fan_show][2] += 5; break;
						// PWM(max)
						case 3: if(settings[mode][fan_show][3] < 255)
									settings[mode][fan_show][3] += 5; break;
					}
					refresh_screen();
					break;
				case 3:							// runter
					if(line == 3)					// unterste Zeile erreicht: Lfter wechseln
						if(fan_show < 6)
						{
							fan_show ++;
							refresh_screen();
						}
						else
						{
							fan_show = 1;
							refresh_screen();
						}
					else							// sonst: eine Zeile runter
					{
						lcd_gotoxy(0, line);
						lcd_putc(' ');
						line ++;
						lcd_gotoxy(0, line);
						lcd_putc(126);	// Pfeil
					}
					break;
				case 4:							// links (-)
					switch(line)
					{	
						// Sensor
						case 0: if(settings[mode][fan_show][0] > 1)
									settings[mode][fan_show][0] --; break;
						// T(ein)
						case 1: if(settings[mode][fan_show][1] > 10)
									settings[mode][fan_show][1] -= 5; break;
						// T(max)
						case 2: if(settings[mode][fan_show][2] > (settings[mode][fan_show][1] + 5))
									settings[mode][fan_show][2] -= 5; break;
						// PWM(max)
						case 3: if(settings[mode][fan_show][3] > PWMmin)
									settings[mode][fan_show][3] -= 5; break;
					}
					refresh_screen();
					break;
			}
		}
		else if((button_counter == 40) && (button == 5))	// speichern
		{
			eeprom_write_block(&settings[mode][fan_show][0], &ee_settings[mode][fan_show][0], 4);	// Lftereinstellungen im EEPROM ablegen

			lcd_clrscr();
			lcd_puts("Lfter");				// 1.Zeile: "Lfter1"
			lcd_putc(fan_show + 48);
			lcd_gotoxy(3,2);
			lcd_puts("Einstellungen");		// 3.Zeile: "Einstellungen"
			lcd_gotoxy(4,3);
			lcd_puts("gespeichert!");		// 4.Zeile: "gespeichert!"
			for(i=0; i<80; i++)				// 2000 ms warten
				_delay_ms(25);
			print_screen();
		}
		else					// keine Taste ausgewertet
			_delay_ms(20);		// 20 ms warten
		
		// Temperatur auslesen und auswerten
		if( !(ADCSRA & (1<<ADSC) ) )		// wenn letzte Konversation beendet
		{
			adc_value = ADC;				// Ergenis der letzten Konversation sichern
			
			// adc_value in Temperatur umrechnen
			temp_sens[sens_read] = (adc_value * 11) / 100;				// linearisierte Temeraturkurve (nur im Bereich -20...70C realistisch)
			
			// Temperatur in LSB(ADC) umrechnen (entspricht ADC-Wert)
			Tmin = ((settings[mode][fan_read][1] * 100) + 50) / 11;	// +50 vermeidet zu kleine Werte durch abschneiden der Kommastellen
			Tmax = ((settings[mode][fan_read][2] * 100) + 50) / 11;	// Ergebnis liegt genau in der Mitte aller zur Ausgangstemp. fhrenden Werte

			// Lfter regeln
			if(adc_value >= Tmax)										// T >= T(max)
				* PWM_register[fan_read] = settings[mode][fan_read][3];
			else if	(adc_value >= Tmin)									// T >= T(min)
			{
				if(*PWM_register[fan_read] != 0)						// lfter luft bereits
				{
					// PWM-Wert errechnen
					*PWM_register[fan_read] = ( ((adc_value - Tmin) * (settings[mode][fan_read][3] - PWMmin)) / (Tmax - Tmin) ) + PWMmin;
				}
				else													// Lfter luft noch nicht, also starten
				{
					*PWM_register[fan_read] = 180;
					for(i=0; i<80; i++)		// 2000 ms warten
						_delay_ms(25);
				}
			}
			else if((adc_value < (Tmin - 10)) && (*PWM_register[fan_read] != 0))	// Hysterese (1 K) unterschritten und Lfter luft
			{
				*PWM_register[fan_read] = 0;							// Lfter deaktivieren
			}
			
			if(fan_read < 6)				// nchsten Lfter anwhlen
				fan_read ++;
			else
			{
				fan_read = 1;
				quick_refresh_screen();		// dynamische Werte aktualisieren nachdem alle ausgelesen wurden
			}
			sens_read = settings[mode][fan_read][0];
			ADMUX = (63 + sens_read);		// den zum aktuellen Lfter gehrenden Sensor anwhlen
			ADCSRA |= (1<<ADSC);			// neue Konversation starten
		}
	}

	return 0;
}

/*
ToDO:
weitere Anzeige: bersicht (Temp, PWM-Wert)
	wie lsen?

starten ohne Display mglich!?! (vor Initialisierung: Antwort von Display
abwarten und gegebenenfalls alle Display-Ausgaben abbrechen)
*/